# Copyright 2012, 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import textwrap
import yaml

from charmworld.models import (
    Bundle,
    Charm,
    CharmSource,
    QAData,
    QADataSource,
    store_bundles,
)
from charmworld import qa_questions
from charmworld.search import (
    BUNDLE,
    IndexMissing,
)
from charmworld.testing import (
    factory,
    MongoTestBase,
)
from charmworld.migrations import migrate
from charmworld.migrations.migrate import Versions

import mock


class MigrationTestBase(MongoTestBase):

    def setUp(self):
        super(MigrationTestBase, self).setUp()
        self.versions = Versions(migrate.get_migration_path())
        self.use_context(mock.patch.object(migrate, 'configure_logging'))


class TestMigration019(MigrationTestBase):

    def setUp(self):
        super(TestMigration019, self).setUp()
        self.use_index_client()

    def test_new_db_and_index_have_all_old_data(self):
        source = CharmSource.from_request(self)
        charm_1 = factory.get_charm_json(
            name='one',
            provides={
                'website': {
                    'interface': 'http',
                },
            },
            requires={
                'database': {
                    'interface': 'mysql',
                }
            })
        source.save(charm_1)
        charm_2 = factory.get_charm_json(
            name='two',
            provides={
                'database': {
                    'interface': 'mysql',
                }
            })
        source.save(charm_2)

        exodus_update = self.versions.get_exodus(
            '019_fixed_mapping_for_interfaces.py')
        target = self.versions.get_pending_source(source, 19)
        target.index_client.create_index()
        target.index_client.put_mapping()
        self.addCleanup(target.index_client.delete_index)
        exodus_update(source, target, 19)
        self.assertEqual(2, target.collection.find().count())
        self.assertEqual(2, target.index_client.search('mysql')['matches'])
        self.assertEqual(1, target.index_client.search('http')['matches'])


class TestMigration020(MigrationTestBase):

    def setUp(self):
        super(TestMigration020, self).setUp()
        self.use_index_client()

    def test_review_entries_updated(self):
        review_entries = [
            factory.make_review_entry(),
            factory.make_review_entry()]
        review_entries[0]['item'] = 'code.launchpad.net'
        review_entries[1]['item'] = 'bugs.launchpad.net'
        self.db.review_queue.insert(review_entries)
        self.versions.run_migration(
            self.db, self.index_client,
            '020_adds_type_field_to_existing_review_items.py')
        results = list(self.db.review_queue.find())
        self.assertEqual(2, len(results))
        for entry in results:
            if entry['item'] == 'code.launchpad.net':
                self.assertEqual('proposal', entry['type'])
            elif entry['item'] == 'bugs.launchpad.net':
                self.assertEqual('task', entry['type'])
            else:
                self.fail()

    def test_fails_on_unknown(self):
        review_entries = [factory.make_review_entry()]
        review_entries[0]['item'] = 'example.com'
        self.db.review_queue.insert(review_entries)
        self.assertRaises(
            Exception,
            self.versions.run_migration,
            self.db, self.index_client,
            '020_adds_type_field_to_existing_review_items.py')


class TestMigration021(MigrationTestBase):

    def setUp(self):
        super(TestMigration021, self).setUp()
        self.use_index_client()

    def test_bundles_removed(self):
        self.assertEqual(0, self.db.bundles.count())
        names = []
        for i in xrange(5):
            bundle, bundle_data = factory.makeBundle(self.db)
            self.index_client.index_bundle(bundle_data)
            names.append(bundle.name)
        self.assertEqual(5, self.db.bundles.count())
        for name in names:
            results = self.index_client.search(name)['results']
            self.assertEqual(1, len(results[BUNDLE]))
        # Run the migration.
        self.versions.run_migration(
            self.db, self.index_client,
            '021_drop_unproofed_bundles.py')
        # All bundles removed.
        self.assertEqual(0, self.db.bundles.count())
        # Bundles removed from elasticsearch.
        for name in names:
            results = self.index_client.search(name)['results']
            self.assertEqual(0, len(results[BUNDLE]))


class TestMigration022(MigrationTestBase):

    # Test removal of  QA questions and data.  The new questions will be
    # updated in the next migration.

    def setUp(self):
        super(TestMigration022, self).setUp()
        self.use_index_client()

    def test_question_removal(self):
        # Load question set.
        category = factory.makeQAQuestionGroup()
        self.db.qa.insert(category)
        self.qa_data_source = QADataSource.from_db(self.db)
        # Show that questions exist before running the migration.
        qadata = QAData(self.qa_data_source)
        self.assertListEqual(['reliable'], qadata.categories)
        # Run the migration.
        self.versions.run_migration(
            self.db, self.index_client,
            '022_Remove_all_QA_questions_and_data.py')
        # And verify the data were removed.
        qadata = QAData(self.qa_data_source)
        self.assertListEqual([], qadata.categories)

    def test_charm_score_removal(self):
        charm_data = factory.makeCharm(self.db)[1]
        charm = Charm(charm_data)
        qa_data_source = QADataSource.from_db(self.db)
        qa_data_source.save_qa_data(charm, {'a': 'b'})
        answers = self.db.charm_qa_answers.find()
        self.assertEqual(1, answers.count())
        # Run the migration.
        self.versions.run_migration(
            self.db, self.index_client,
            '022_Remove_all_QA_questions_and_data.py')
        # And verify the data were removed.
        answers = self.db.charm_qa_answers.find()
        self.assertEqual(0, answers.count())


class TestMigration023(MigrationTestBase):

    # Test restoration of new QA questions.

    def setUp(self):
        super(TestMigration023, self).setUp()
        self.use_index_client()

    def test_questions_inserted(self):
        # Load question set.
        self.qa_data_source = QADataSource.from_db(self.db)
        # Show that no questions exist before running the migration.
        qadata = QAData(self.qa_data_source)
        self.assertListEqual([], qadata.categories)
        # Run the migration.
        self.versions.run_migration(
            self.db, self.index_client,
            '023_insert_new_questions.py')
        # And verify the data were removed.
        qadata = QAData(self.qa_data_source)
        self.assertEqual(7, len(qadata.categories))
        expected = [cd['name'] for cd in qa_questions.iter_categories()]
        self.assertListEqual(sorted(expected), sorted(qadata.categories))


class TestExodusRestrictions(MigrationTestBase):

    def test_exodus_restrictions(self):
        """Ensure that deploy-from-scratch does not violate exodus rules."""
        self.versions._check_exodus(self.versions.list_pending(0))
